//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// YASA robust earliest deadline first implementation
//
// Project: Yasa 2
// Author : Jan Blumenthal
// Start  : 2001/11/26
// $Header: /sources/yasa/schedulers/red/red.c,v 1.3 2003/01/23 09:18:33 bj Exp $
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program (See the included file COPYING);
// if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Features: Support for Stack Resource Protocol (SRP)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "red.h"


////////////////////////////////////////////////////////////////////////////////////////////////////////////
YASA_THREAD* red_schedule(YASA_SCHEDULER *scheduler, const YASA_TIME *now, YASA_TIME *ptime)
{
	YASA_THREAD		*thread;
	YASA_THREAD		*rejectthread;
	YASA_THREAD		*newthread;
	int				reschedule;

	do
	{
		yasa_initschedulingcycle(newthread);
		reschedule=0;
		for (thread = YASA_TASKSET(scheduler); thread; thread = YASA_NEXT_THREAD(thread) )
		{
			if ( YASA_CHECK_NEXT_ACTIVATION(thread) )
				yasa_setnextactivationtime(thread, now);

			// Should the thread be (re)started in future?
			// thread->threadflags&TIMER_ARMED
			// The thread should be restarted. Did we reach the resume time?
			if ( YASA_TEST_TIMER_ARMED(thread) && *now >= YASA_RESUME_TIME(thread) )
			{
				// timer is no more armed. thread is ready to compute
				// thread->threadflags&=~TIMER_ARMED
				YASA_UNSET_TIMER_ARMED(thread);

				// thread can be choosen in this scheduling !
				// thread->pending will 'TRUE'
				YASA_SIGNAL_TIMER(thread);

				// set next(!) resume time
				yasa_setdeadline(thread);			// set deadline of current thread
			}

			// find highest priority runnable task
			if (  YASA_SCHEDULABLE(thread)  && !YASA_DEADLINE_ALREADY_MISSED(thread) )
			{
				// if deadline missed, there can be started an emergency thread
				// rescheduling neccessary !!!
				if ( YASA_TEST_DEADLINE_MISSED(thread,now) )
				{
					reschedule|=yasa_deadlinemissed(thread, now);
				}
				else
				{
					// schedule only non rejected threads -> "work queue"
					if ( !newthread || YASA_CURRENT_DEADLINE(thread) < YASA_CURRENT_DEADLINE(newthread) )
					{
#ifdef YASA_SUPPORT_DYNAMICALLY_SYNCHRONIZATION_SCHEDULING
						if ( yasa_cmpdynamicpriority( &YASA_DEADLINE(thread) ) )
#endif
						newthread=thread;
					}
				}
			}
		}
	} while (reschedule);


	// check "REJECT Queue"
	rejectthread=0;
	for (thread = YASA_TASKSET(scheduler); thread; thread = YASA_NEXT_THREAD(thread) )
	{
		if ( YASA_DEADLINE_ALREADY_MISSED(thread) )			// only rejected threads !
		{
			// scheduling algorithm
			if ( YASA_CURRENT_DEADLINE(thread)+YASA_DEADLINE_TOLERANCE(thread) <= *now)
			{
				yasa_deadlinemissed(thread, now);
			}
			else if ( ! newthread )
			{
				if ( !rejectthread || YASA_CURRENT_DEADLINE(thread)+YASA_DEADLINE_TOLERANCE(thread) < YASA_CURRENT_DEADLINE(rejectthread) + YASA_DEADLINE_TOLERANCE(thread) )
				{
#ifdef YASA_SUPPORT_DYNAMICALLY_SYNCHRONIZATION_SCHEDULING
					if ( yasa_cmpdynamicpriority( &YASA_DEADLINE(thread) ) )
#endif
						rejectthread=thread;
				}
			}
		}
	}

	if ( !newthread )
		newthread=rejectthread;


	/////////////////////////////////////////////////////////////////////////////////////////
	// determine preemption time. Please pay attention to:
	// 1. should the tick or the preemption time be used
	// 2. what happens if newthread=0 ?
	// 3. what happens with non computing threads having deadlines until *ptime ?
	// 4. You cannot be sure to get valid values if the project does not support this scheduler
	//    Example: In priority based schedulers the priority should be analyzed only?
	if ( YASA_CALCPREEMPTIONTIME(scheduler) )
	{
		*ptime=YASA_TIME_INFINITY;
		for (thread = YASA_TASKSET(scheduler); thread; thread = YASA_NEXT_THREAD(thread) )
		{
			if ( YASA_TEST_TIMER_ARMED(thread) )
			{
				// Here could be calculated the laxity of thread and newthread at resume time
				// and could be compared but I think that is will cost more time than a new scheduling!?
				if( ! YASA_COMPUTED_TIME(thread) )	// if computed time=0 -> current deadline is not set and the thread will be started in future
				{
					if ( !newthread || YASA_DEADLINE(thread)+YASA_RESUME_TIME(thread) < YASA_CURRENT_DEADLINE(newthread) )
						if ( *ptime > YASA_RESUME_TIME(thread) )
							*ptime=YASA_RESUME_TIME(thread);
				}
				else								// current deadline is set, maybe thread is waiting for something
					if ( !newthread || YASA_CURRENT_DEADLINE(thread) < YASA_CURRENT_DEADLINE(newthread) )
						if ( *ptime > YASA_RESUME_TIME(thread) )
							*ptime=YASA_RESUME_TIME(thread);
			}
		}
	}
	else
		YASA_SETTICKSCHEDULINGTIME(ptime, now, scheduler)
		//*ptime = *now + scheduler->tick;

	return newthread;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int red_init(YASA_SCHEDULER *scheduler)
{
	return yasa_setdynamicprioritytype(YASA_SYNC_DEADLINE);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// static definitions
const YASA_SCHEDULER_ATTR red_attr=
{
	0,						// next
	{ 0, "ys_red" },		// id & name
	RED_VERSION,			// scheduler version
	0,						// userdata
	red_init,				// init()
	0,						// cleanup()
	red_schedule,			// schedule()
	0,						// findpreemptor()
};
