//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// YASA logfile parser
//
// Project: Yasa 2
// Author : Jan Blumenthal
// Start  : 2002/02/28
// $Header: /sources/yasa/yasagui/logfile.h,v 1.4 2003/01/24 15:47:41 bj Exp $
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program (See the included file COPYING);
// if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef YASAGUI_LOGFILE_PARSER_INCLUDE
#define	YASAGUI_LOGFILE_PARSER_INCLUDE

#include <yasagui/yasagui.h>
#include "datatypes.h"
#include "list.h"
#include "filedialog.h"
#include "logfilestatistics.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YProject;
class YCPULogFile;
class YCPULogFileSet;
class YLogAction;
class YLogFileTask;
class YLogFileMutex;
class YEnvironmentLogFileSet;
class YProjectLogFileSet;

typedef unsigned char		ACTIONINT;
#define ACTIONINTMAX		255

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YLoggingParser
{
	YASA_TIME				starttime;								// current parsing start time
	YASA_TIME				endtime;								// current parsing end time
	//! ptr to the current cpulogfile set (represents one environment)
	YCPULogFile				*cpulogfile;
	int						cpulogfilenumber;
	//! ptr to current environment logfile (represents all cpus in the current environment)
	YEnvironmentLogFileSet	*environmentlogfile;
	int						environmentlogfilenumber;
	//! ptr to current project logfile (represents all environment logfiles in the project)
	YProjectLogFileSet		*projectlogfile;
	int						projectlogfilenumber;
public:						YLoggingParser() :
							starttime(0),
							endtime(0),
							cpulogfile(0),
							cpulogfilenumber(0),
							environmentlogfile(0),
							environmentlogfilenumber(0),
							projectlogfile(0),
							projectlogfilenumber(0)
							{}
	//! Set a new timing interval where to serach for logging events
	void					SetTime(YASA_TIME start, YASA_TIME end)		{	starttime=start;
																			endtime=end;						}
	YASA_TIME				GetStart() const							{	return starttime;					}
	YASA_TIME				GetEnd() const								{	return endtime;						}
	virtual bool			TimingFilter(const YLogAction *) const		{	return true;						}
	virtual void			EvaluateLoggingEntry(YLogAction *)=0;

	YCPULogFile	*			GetCPULogFile()	const						{	return cpulogfile;					}
	int						GetCPULogFileNumber() const					{	return cpulogfilenumber;			}
	void					SetCPULogFile(YCPULogFile *t, int number)	{	cpulogfile=t;
																			cpulogfilenumber=number;			}
	YEnvironmentLogFileSet*	GetEnvironmentLogFile() const				{	return environmentlogfile;			}
	int						GetEnvironmentLogFileNumber() const			{	return environmentlogfilenumber;	}
	void					SetEnvironmentLogFile(YEnvironmentLogFileSet *t, int number)
																		{	environmentlogfile=t;
																			environmentlogfilenumber=number;	}
	void					ResetParser()								{	cpulogfile=0;
																			cpulogfilenumber=0;
																			environmentlogfile=0;
																			environmentlogfilenumber=0;
																			projectlogfile=0;
																			projectlogfilenumber=0;
																		}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class is used to store information about one logging entry. If there is a more comlplex entry
//! like a start of a thread a derived class YThreadLogAction/YThreadRangeLogAction is used.
//! Normally the API would be designed with virtual functions. In these classes no virtual function are used
//! to minimize the usage of memory.
class YLogAction
{
	YASA_TIME				starttime;
	YASA_TIME				endtime;
	YLogFileTask *			thread;						// ptr to thread, if this logging entry is using a thread
	YLogFileMutex *			mutex;						// ptr to mutex, if this logging entry is using a mutex
	// YASA_IDLE_ID			-1
	// YASA_SCHEDULER_ID	-2
	// YASA_MUTEX_ID		-3	// Startup code
	// YASA_UNDEFINED		-4
	ACTIONINT				action;						// thread started, deadline missed etc.
public:						YLogAction() :
							starttime(0),
							endtime(0),
							thread(0),
							mutex(0),
							action(YASAL_UNKNOWN)
							{}
	QString					DetailedInfo();
	YASA_TIME				GetStartTime() const			{	return starttime;							}
	YASA_TIME				GetEndTime() const				{	return endtime;								}
	YLogFileTask *			GetThread() const				{	return thread;								}
	YLogFileMutex *			GetMutex() const				{	return mutex;								}
	ACTIONINT				GetAction() const				{	return action;								}
	void					SetStartTime(const YASA_TIME &t){	starttime=t;
																endtime=t;									}
	void					SetEndTime(const YASA_TIME &t) 	{	endtime=t;									}
	void					SetThread(YLogFileTask *t)		{	thread=t;									}
	void					SetMutex(YLogFileMutex *m)		{	mutex=m;									}
	void					SetAction(ACTIONINT a)			{	action=a;									}
	bool					IsThreadRange() const			{	return !mutex && endtime;				 	}
	bool					IsMutexRange() const			{	return mutex && endtime;					}
	bool					IsExcludedThreadRange() const	{	return !mutex && endtime && action==YASAL_THREAD_EXCLUDED;	}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! Properties of a task extracted from logfile
class YLogFileTask : public YObject
{	int						id;												// can be 1,2,3,4 -1,-2-3
	int						number;											// number in list
	int						periods;
	YTaskLogFileStatistics	statistics;
	// temporary values, used to determine statistic values
	YASA_TIME				thread_activated;
	YASA_TIME				thread_computed;
	YASA_TIME				thread_scheduledactivationtime;
public:						YLogFileTask(const YObjectArgs &args) : YObject(args),
							id(0), number(0), periods(0), thread_activated(0), thread_computed(0), thread_scheduledactivationtime(0)
																			{	statistics.SetTask(this);		}
	int						GetId() const									{	return id;						}
	void					SetId(int i)									{	id=i;							}
	int						GetNumber() const								{	return number;					}
	void					SetNumber(int i)								{	number=i;						}
	YTaskLogFileStatistics&	GetStatistics()									{	return statistics;				}
	void					SetThreadActivatedTmp(YASA_TIME t)				{	thread_activated=t;				}
	void					SetScheduledActivationTime(YASA_TIME t)			{	thread_scheduledactivationtime=t;}
	void					SetThreadPeriod(YASA_TIME currenttime)
/*															if ( GetStatistics().GetPeriod()==YASA_TIME_INFINITY
																 && currenttime>thread_scheduledactivationtime )
															{
																GetStatistics().SetPeriod( currenttime - thread_scheduledactivationtime);
															}
															thread_scheduledactivationtime=0;
														}
*/							{
								if ( periods )
								{
									GetStatistics().SetPeriod( currenttime - thread_scheduledactivationtime);
								}
								thread_scheduledactivationtime=currenttime;
								periods++;
							}
	void					SetThreadDeadline(YASA_TIME currentdeadline)
/*							{	if ( currentdeadline>thread_scheduledactivationtime
									 && GetStatistics().GetDeadline()==YASA_TIME_INFINITY )
								{
									GetStatistics().SetDeadline( currentdeadline - thread_scheduledactivationtime);
								}
							}
*/
							{
								if ( periods && currentdeadline>thread_scheduledactivationtime )
								{
									GetStatistics().SetDeadline( currentdeadline - thread_scheduledactivationtime);
								}
							}


	void					SetThreadComputedTmp(YASA_TIME t)				{	thread_computed=t;				}
	void					AddThreadComputedTmp(YASA_TIME t)				{	thread_computed+=t;				}
	YASA_TIME				GetThreadActivatedTmp() const					{	return thread_activated;		}
	YASA_TIME				GetThreadComputedTmp()	const					{	return thread_computed;			}
	bool					IsIdleTask() const								{	return GetId() == YASA_IDLE_ID;	}
	bool					IsSystemTask() const							{	return GetId() == YASA_SCHEDULER_ID;}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! List of all tasks of one logfile
class YLogFileTaskSet : public YSet
{ protected:
	virtual YObject*		AllocNewItem(const YObjectArgs &args)			{	return new YLogFileTask(args);	}
public:						YLogFileTaskSet() :
							YSet( 0, MSGP_LF_DEFAULT_TASKNAME, YObjectArgs(), 0 )	{ }
	YLogFileTask *			AllocTaskItemOnce(const QString &name, int threadid)
							{	YLogFileTask	*task=FindTask(threadid);
								if (task)
									return task;
								if (name.length() )
									task = (YLogFileTask*) CreateNewItem(name);
								else
									task = (YLogFileTask*) CreateNewItem(MSGP_LF_DEFAULT_TASKNAME.arg(threadid) );
								if (task)
								{
									task->SetId(threadid);
									task->SetNumber(GetCount()-1);
								}
								return task;
							}
	YLogFileTask *			FindTask(int threadid)
							{
								for( unsigned int i=0; i<GetCount() ; i++)
								{	if ( ((YLogFileTask*)At(i))->GetId() == threadid )
										return (YLogFileTask*)At(i);
								}
								return 0;
							}
	void					SortTaskSet()									{	sort();							}
	void					ResetStatistics()
							{
								for( unsigned int i=0; i<GetCount() ; i++)
									((YLogFileTask*)At(i))->GetStatistics().ResetStatistics();
							}
	void					GenerateStatistics(YEnvironmentLogFileSet &envlogset)
							{
								for( unsigned int i=0; i<GetCount() ; i++)
									((YLogFileTask*)At(i))->GetStatistics().CalculateStatistics( envlogset );
							}
protected:	int				compareItems ( QCollection::Item item1, QCollection::Item item2 )
							{	if ( ((YLogFileTask*)item1)->GetId() < ((YLogFileTask*)item2)->GetId() )
									return -1;
								if ( ((YLogFileTask*)item1)->GetId() == ((YLogFileTask*)item2)->GetId() )
									return 0;
								return 1;
							}
};
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! Properties of a mutex extracted from a logfile
class YLogFileMutex : public YObject
{	int						id;
	int						number;							// same that results by calling yset.at()
	YMutexLogFileStatistics	statistics;
	YASA_TIME				lockingtime;
public:						YLogFileMutex(const YObjectArgs &args) : YObject(args), id(0), number(0), lockingtime(0)
																			{	statistics.SetMutex(this);		}
	int						GetId() const									{	return id;						}
	void					SetId(int i)									{	id=i;							}
	int						GetNumber() const								{	return number;					}
	void					SetNumber(int n)								{	number=n;						}
	YMutexLogFileStatistics&GetStatistics()									{	return statistics;				}
	void					SetMutexLockingTimeTmp(YASA_TIME t)				{	lockingtime=t;					}
	//void					AddMutexLockingTimeTmp(YASA_TIME t)				{	lockingtime+=t;					}
	YASA_TIME				GetMutexLockingTimeTmp() const					{	return lockingtime;				}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! List of all mutexes of one logfile
class YLogFileMutexSet : public YSet
{ protected:
	virtual YObject*		AllocNewItem(const YObjectArgs &args)			{	return new YLogFileMutex(args);	}
public:						YLogFileMutexSet() :
							YSet( 0, MSGP_LF_DEFAULT_TASKNAME, YObjectArgs(),0 )	{ }
	YLogFileMutex *			AllocMutexItemOnce(const QString &name, int mutexid)
							{	YLogFileMutex	*mutex;
								for( unsigned int i=0; i<GetCount() ; i++)
								{	if ( ((YLogFileMutex*)At(i))->GetId() == mutexid )
										return (YLogFileMutex*)At(i);
								}
								if ( name.length() )
									mutex = (YLogFileMutex*)CreateNewItem( name );
								else
									mutex = (YLogFileMutex*)CreateNewItem( MSGP_LF_DEFAULT_MUTEXNAME.arg(mutexid) );
								if (mutex)
								{
									mutex->SetId(mutexid);
									mutex->SetNumber(GetCount()-1);
								}
								return mutex;
							}
	void					SortMutexSet()									{	sort();							}
	void					ResetStatistics()
							{
								for( unsigned int i=0; i<GetCount() ; i++)
									((YLogFileMutex*)At(i))->GetStatistics().ResetStatistics();
							}
	void					GenerateStatistics(YEnvironmentLogFileSet &envlogset)
							{
								for( unsigned int i=0; i<GetCount() ; i++)
									((YLogFileMutex*)At(i))->GetStatistics().CalculateStatistics( envlogset );
							}
protected:	int				compareItems ( QCollection::Item item1, QCollection::Item item2 )
							{	if ( ((YLogFileMutex*)item1)->GetId() < ((YLogFileMutex*)item2)->GetId() )
									return -1;
								if ( ((YLogFileMutex*)item1)->GetId() == ((YLogFileMutex*)item2)->GetId() )
									return 0;
								return 1;
							}
};



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! Logfile File descriptor
class YLogFile : public YFile
{	int						linenumber;
public:						YLogFile(const QString &f) :
							YFile(f),
							linenumber(0)
							{}
	int						ReadLine(QString &);
	int						GetLineNumber() const					{	return linenumber;	}
	int						ReadGlobalArg(YASA_TIME &time, const QString &comparestring);
	int						ReadGlobalDoubleArg(double &value, const QString &comparestr );
};



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This list contains logging entries of one CPU in an environment
class YLogActions : public QList<YLogAction>
{	public:					YLogActions()
							{
								setAutoDelete(TRUE);
							}
	YLogAction*				InsertNewLogAction(YLogAction *log)
							{
								if ( log )
									append(log);
								return log;
							}
	void					RemoveLogAction(YLogAction *log)
							{
								if ( log )
									removeRef(log);
							}
	// The current item of the list is used to traverse the list
	void					EvaluateLoggingActions(YLoggingParser &p)
							{
								if ( ! count() )													// return if no item in list
									return;
								if ( !first() )
									return;
								while ( at()!=-1 && current()->GetStartTime()<p.GetEnd() )
								{
									if ( current()->GetStartTime() <= p.GetEnd() &&
										 current()->GetEndTime() >= p.GetStart() )				// go forward in list to first item after start time
									{
										p.EvaluateLoggingEntry( current() );
									}
									next();
								};
								return;
							}
protected:	int				compareItems ( QCollection::Item item1, QCollection::Item item2 )
							{	if ( ((YLogAction*)item1)->GetStartTime() < ((YLogAction*)item2)->GetStartTime() )
									return -1;
								if ( ((YLogAction*)item1)->GetStartTime() == ((YLogAction*)item2)->GetStartTime() )
									return 0;
								return 1;
							}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! Logging properties of one CPU
class YCPULogFile : public YObject
{	Q_OBJECT
	YLogActions				logactions;
	YCPULogFileStatistics	statistics;
public:						YCPULogFile(const YObjectArgs &i) :	YObject(i)	{ }
							~YCPULogFile()
							{
								emit RemoveLogFile();
							}
	//! Parse CPU logfile of file and create logfile instances
	int						ParseCPUSection(YLogFile &file);
	int						GetCPU() const									{	return statistics.GetNumberOfCPU()-1;	}
	void					SetCPU(int cpu)									{	statistics.SetNumberOfCPU(cpu+1);		}
	//! Returns the max time of this cpu
	YASA_TIME				GetDuration() const								{	return statistics.GetExecutionDuration();}
	YCPULogFileStatistics &	GetStatistics()									{	return statistics;						}
	YLogActions &			GetLogActions()									{	return logactions;						}
	YLogAction*				CreateJitterLogAction(YASA_TIME time, YASA_TIME endtime);
	static void				SetEndRange(YLogAction **range, YLogAction *action, YASA_TIME t);
	static YASA_TIME		ReadYasaTimeParam(const QString &str, bool &ok);
	static QString			ExtractName(const QString src);
signals:
	void					RemoveLogFile();
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! A YEnvironmentLogFileSet contains data of one environment logfile
class YEnvironmentLogFileSet : public YSet
{	Q_OBJECT
	YLogFileTaskSet			taskset;
	YLogFileMutexSet		mutexset;
	QDateTime				readdate;
	YEnvironmentLogFileStatistics statistics;
public:						YEnvironmentLogFileSet();
protected:
	virtual YObject*		AllocNewItem(const YObjectArgs &args)			{	return new YCPULogFile(args);	}
public:
	YLogFileTaskSet &		GetTaskSet()									{	return taskset;					}
	YLogFileMutexSet &		GetMutexSet()									{	return mutexset;				}
	void					ResetEnvironmentLogFile();
	void					ResetStatistics();
	int						GenerateStatistics();
	int						ParseLogFile(const QString &filename);
							~YEnvironmentLogFileSet()						{	CleanupLogFile();				}
	//! Cleaning up logfile buffers. All internal logfiles will be detached. The logfiles will not be removed from hard disk
	void					CleanupLogFile()
							{
								if ( GetCount() || taskset.GetCount() || mutexset.GetCount() )
								{
									emit LogFileCleanedUp();
									ResetEnvironmentLogFile();
									emit LogFileUpdated();
								}
							}
	//! Returns the max time of all parsed cpu logfiles
	YASA_TIME				GetMaxDuration()
							{
								YASA_TIME	time=0;
								for( unsigned int index=0; index<GetCount();index++)
								{
									if (time< ((YCPULogFile*)At(index))->GetDuration() )
										time=((YCPULogFile*)At(index))->GetDuration();
								}
								return time;
							}

	//! This function parses all logfile of the environment between YLoggingParser::Start until YLoggingParser::End
	//! and is calling the YPainter->DrawContents() function
	void					EvaluateLoggingActions(YLoggingParser &p)
							{	for( unsigned int index=0; index<GetCount();index++)
								{
									p.SetCPULogFile( (YCPULogFile*)At(index), index );
									((YCPULogFile*)current())->GetLogActions().EvaluateLoggingActions( p );
								}
							}
	YEnvironmentLogFileStatistics &	GetStatistics()								{	return statistics;		}
signals:
	//void					RemoveLogFiles();
	void					LogFileUpdated();
	void					LogFileCleanedUp();
};



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! A YProjectLogFileSet contains data of all logfiles belonging to all environments of the project
class YProjectLogFile : public YObject
{	//Q_OBJECT
	YProject				*project;
	void					UpdateList();
public:						YProjectLogFile() : project(0)				{							}
	void					EvaluateLoggingActions(YLoggingParser &p);
	//! Returns the max time of all parsed cpu logfiles
	void					SetParent(YProject *p)						{	project=p;				}
	YProject*				GetProject() const							{	return project;			}
	YASA_TIME				GetMaxDuration() const;
	unsigned int			GetMaxEntries() const;
	QStringList				GetEnvironmentNames() const;
/*
public slots:
	//! A new environment was created. The logfile set must be updated
	void					NewEnvironment(YObject *object)				{	AppendItem(object);		}
	//! An environment was removed. The logfile set must be updated
	void					EnvironmentRemoved(YObject *object)			{	RemoveItem(object);		}
*/
};












































//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class LogFileParser
{
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YasaLogFileParser
{
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class OldYasaLogFileParser
{
};




//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif																	// ifndef YASAGUI_LOGFILE_PARSER_INCLUDE
