//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// YASA application config
//
// Project: Yasa 2
// Author : Jan Blumenthal
// Start  : 2002/02/28
// $Header: /sources/yasa/yasagui/object.h,v 1.6 2003/01/24 15:47:41 bj Exp $
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program (See the included file COPYING);
// if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef YASAGUI_OBJECT_INCLUDE
#define	YASAGUI_OBJECT_INCLUDE


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <yasagui/yasagui.h>
#include <qobject.h>
#include "translations.h"
#include "config.h"


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YInputParser;
class YOutputParser;
class YActions;
class YSet;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class contains the arguments to create an object. The contents will be copied during
//! initialization so it is safe to destroy YObjectArgs later.
class YObjectArgs
{	//! localized name of object e.g. "task 1" if set otherwise ""
	QString				name;
	//! ptr to object actions
	YActions			*actions;
	//! ptr to parent list which contains the object to be created
	YSet				*set;
public:
						//@{
						//! These are the constructors of YID. The instance will be
						//! initialized with the given parameters. After construction
						//! of the final object it is safe to destroy the objectargs
						YObjectArgs( const YObjectArgs &i, YActions *a=0, YSet *s=0 ) :
						name(i.name),
						actions(a),
						set(s)
						{}
						YObjectArgs( const QString &str=QString::null, YActions *a=0, YSet *s=0 ) :
						name(str),
						actions(a),
						set(s)
						{}
						//@}
	//! Returns the name of the object
	const QString&		GetName() const				{	return name;								}
	//! Returns a ptr to the global actions. YActions* can be zero.
	YActions*			GetActions() const			{	return actions;								}
	//! Returns a ptr to the list if this object contains to a list. YSet* can be zero
	YSet*				GetSet() const				{	return set;									}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is a base definition for every list entry in the appplication. The name will be stored
//! in base class "QString"
class YObject : public QObject, protected QString
{	Q_OBJECT
	//! Contains a description of this object
	QString				description;
	//! Contains a ptr to a YSet if used.
	YActions			*actions;
	//! link to YSet, can be zero if not in a list
	YSet				*set;
	//! Contains information about the state of the object (true=changed, false=unchanged)
	bool				changed;
public:					YObject(const YObjectArgs &i=YObjectArgs() ) :
						QString( i.GetName() ),
						actions( i.GetActions() ),
						set( i.GetSet() )
						{}
						/*~YObject()
						{
							//emit DestroyObject();
							//YActions::ObjectDestroyed(this);
						}*/
	//! Sets the internal flags that signs this object is changed
	void				SetChanged(bool b)			{	changed=b;								}
	//\return true if the object flag is set.
	bool				GetChanged() const			{	return changed;							}
	//! Returns the name of the object
	const QString&		GetName() const				{	return *this;							}
	//! Returns the description of the object
	const QString&		GetDescription() const		{	return description;						}
	//! Sets the description of the object
	//\return Error or NULL
	int					SetDescription(const QString &d);
	//! Sets the name of the object
	//\return Error or NULL
	int					SetName(const QString &n);
	//! Sets the name of the object
	//\return Error or NULL
	int					SetName(const char *n);
	virtual int			SetConfig(YInputParser &parser);
	virtual int			GetConfig(YOutputParser &parser);
	//! Returns a ptr to the global actions. YActions* can be zero.
	YActions*			GetActions() const			{	return actions;							}
	//! Returns a ptr to the list if this object contains to a list. YSet* can be zero
	YSet*				GetSet() const				{	return set;								}
	//! Connect the object with the given one. The "member" function must not contain arguments
	void				UpdateConnect(const QObject *o, const char *member) const
						{
							connect( this, SIGNAL(UpdateObject() ), o, member);
						}
	//! Connect the object with the given one. The "member" function can only contain the argument "YObject *"
	void				UpdateArgConnect(const QObject *o, const char *member) const
						{
							connect( this, SIGNAL(UpdateObject(YObject *) ), o, member);
						}
signals:
	//! Signals that this object was changed
	void				UpdateObject();
	//! Signals that this object was changed
	void				UpdateObject(YObject *);
public slots:
	//! Will be called if the current object or derived objects are changed
	void				ObjectUpdated()				{	changed=true;
														emit UpdateObject();
														emit UpdateObject(this);				}
};




//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is a base class for YDerivedType<T> to support signal/slot mechanism of QT
class YDerivedObject : public YObject
{	Q_OBJECT
protected:				YDerivedObject(const YObjectArgs &args) :
						YObject(args)
						{}
	//! connect this object to it's parent. It will be informed. If the parent is changed.
	//! Mostly used in derived object.
	void				ConnectDerivedParent(const YObject *p)
						{	if ( p )
							{
								connect(p, SIGNAL( UpdateObject() ), this, SLOT( DerivedParentContentChanged() ) );
								connect(p, SIGNAL( DisableObjects(bool) ), this, SLOT( ObjectDisabled(bool) ) );
							}
						}
	void				DisconnectDerivedParent(const YObject *p)
						{	if ( p )
							{
								disconnect(p, SIGNAL( UpdateObject() ), this, SLOT( DerivedParentContentChanged() ) );
								disconnect(p, SIGNAL( DisableObjects(bool) ), this, SLOT( ObjectDisabled(bool) ) );
							}
						}
protected:
	//! This function is called to inform all child objects due to emitting the Signal
	//! DisableObjects() to enable/disable their content.
	void				DisableChildObjects(bool b)				{	emit DisableObjects(b);			};
public:
	//! Returns the content as string, e.g. for writing configuration files
	virtual QString		GetContentString()	const				{	return QString::null;			}
	//! Returns a visible string of the content. This is used to print
	//! the contents in a localized way (such as "infinity")
	virtual QString		GetVisibleContentString() const			{	return QString::null;			}
	//! Read the contents from a string, e.g. reading from configuration files
	virtual	int			ReadFromString(const QString &s)		{	return 0;
																	Q_CONSTUNUSED(s);				}
	//! Returns true if this instance has a valid content
	virtual bool		IsValid() const=0;
protected slots:
	//! This is a slot in the base class (tricky right?) to ensure that derived objects
	//! can receive signals from their parent objects.
	//! This functions is called every time an parent object is changed. This function will refer the
	//! signal to its children. Due to possible changes in "object limits" we call SetContent() in
	//! this function.
	virtual void		DerivedParentContentChanged()			{	/*qWarning("Not used by this YObject");*/	}
	//! If the parent object has called the function SetChildsDisabled() all child objects
	//! will receive the signal DisableObjects() in this slot. The content of all child
	//! objects will be disabled.
	virtual void		ObjectDisabled(bool)					{												}
signals:
	//! This signal is emitted during YDerivedType<T>::SetChildsDisabled() to disable all child objects
	void				DisableObjects(bool);
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is a class providing "an object hierarchy" and supports derived objects properties.
/*! This is a template class which can be used with most common datatypes (bool, int, long, long long).
    The content of the derived class can be set using SetContent() or the appropriate operators.
    YDerivedType<T>::operator <<() or YDerivedType<T>::operator >>().
    If valid is set to FALSE the content of this instance is invalid and must not be read! In this
    case the content of the parent instance is used and so on.
    The instance with no parent cannot be invalid!
    \example yasagui/doc/examples/derived_datatypes.cpp
    \include yasagui/doc/examples/derived_datatypes.cpp
**/
class DerivedTest
{
	int	a;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is a class providing "an object hierarchy" and supports derived objects properties.
/*! This is a template class which can be used with most common datatypes (bool, int, long, long long).
    The content of the derived class can be set using SetContent() or the appropriate operators.
    YDerivedType<T>::operator <<() or YDerivedType<T>::operator >>().
    If valid is set to FALSE the content of this instance is invalid and must not be read! In this
    case the content of the parent instance is used and so on.
    The instance with no parent cannot be invalid!
    \example yasagui/doc/examples/derived_datatypes.cpp
    \include yasagui/doc/examples/derived_datatypes.cpp
**/
template<class T>
class DerivedTest2
{
	T			content;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is a class providing "an object hierarchy" and supports derived objects properties.
/*! This is a template class which can be used with most common datatypes (bool, int, long, long long).
    The content of the derived class can be set using SetContent() or the appropriate operators.
    YDerivedType<T>::operator <<() or YDerivedType<T>::operator >>().
    If valid is set to FALSE the content of this instance is invalid and must not be read! In this
    case the content of the parent instance is used and so on.
    The instance with no parent cannot be invalid!
    \example yasagui/doc/examples/derived_datatypes.cpp
    \include yasagui/doc/examples/derived_datatypes.cpp
**/
template<class T>
class YDerivedType : public YDerivedObject
{	/*! This template type contains the element
	//\example yasagui/doc/examples/derived_datatypes.cpp
	**/
	T					content;
	/** true if <content> is valid
	//\include yasagui/doc/examples/derived_datatypes.cpp
	**/
	bool				valid;
	//! true if this object is disabled. In this case parent objects are used. Root objects can not be disabled.
	bool				disabled;
	//! If set all child are disabled to ensure that this property is used in all childs
	bool				childsdisabled;
	//! Ptr to parent of object
	const YDerivedType<T> *derivedparent;
public:					YDerivedType(const T &c, const YDerivedType<T> *p=0, const YObjectArgs &args=YObjectArgs() ):
						YDerivedObject(args),
						//content(c),
						valid(true),
						disabled(false),
						childsdisabled(false),
						derivedparent(0)	// will be set later !
						{
							SetContent(c);
							if (p)
							{
								disabled=p->IsDisabled();
								SetParent(p);
							}
						}
						//! value must not be set. Only parent is valid
						YDerivedType(const YDerivedType<T> *p, const YObjectArgs &args=YObjectArgs() ):
						YDerivedObject(args),
						//content(0),	// not possible due to using types of structs or classes
						valid(false),
						disabled(false),
						childsdisabled(false),
						derivedparent(0)	// will be set later !
						{
							if (!p)
								qWarning("YDerivedType::YDerivedType() called with illegal parent!");
							else
							{
								//disabled=p->AreChildsDisabled();
								SetParent(p);
								//ConnectDerivedParent(p);
							}
						}
						//! This is a Copy-Construktur for YDerivedType. It is not the final version
						//! Use it with care! The contents will be copied mostly from <p> without the
						//! ptr to YSet
						//! Attention: The copied object is from the parents of <p> derived too! So
						//! be sure that the parents are not destroyed before the childs are destroyed!
						//! This is important by using local attributes
						YDerivedType(const YDerivedType<T> &p) :
						YDerivedObject( YObjectArgs(p.GetName(), p.GetActions(), 0 ) ),	// copied instances have not the same set !
						valid( p.IsValid() ),
						disabled( p.IsDisabled() ),
						childsdisabled( p.childsdisabled ),
						derivedparent( 0 ) //will be set later p.GetDerivedParent() )
						{
							if ( p.GetDerivedParent() )
							{
								//ConnectDerivedParent( derivedparent );
								SetParent( p.GetDerivedParent() );
							}
							if ( p.IsValid() )
								SetContent( p.GetContent() );
							qWarning("YDerivedType::YDerivedType(const YDerivedType<T> &p): Use copy constructors with care!");
						}
						//! The Copy-Construktur is used here capture wrong code
						//! Copy constructor are not allowed because it cannot set the correct parent?
						/*YDerivedType(const YDerivedType<T> &p) :
						YDerivedObject( YObjectArgs(p.GetName(), p.GetActions(), 0 ) ),	// copied instances have not the same set !
						//content(p.content)
						//valid(p.set),
						//parent(p.parent)
						{
							qWarning("YDerivedType::YDerivedType(const YDerivedType<T> &p): Use Copy constructors with care!");
							Q_CONSTUNUSED(p);
						}*/
						/*
						~YDerivedType()
						{
							if ( className() == "YASA_PRIORITY" )
								qWarning("YDerivedType::YDerivedType(const YDerivedType<T> &p): Use Copy constructors with care!");
						}*/
	//! Set a new parent. The content will not be infected
	//! parents to zero are illegal. Use Detach instead !!!
	//! A signal "objectdisabled(true|false) is emitted if necessary.
	void				SetParent(const YDerivedType<T> *p)
						{
							if ( p )
							{
								if ( derivedparent )
								{
									DisconnectDerivedParent(derivedparent);
								}
								derivedparent=p;
								ConnectDerivedParent(derivedparent);
								//if ( derivedparent->AreChildsDisabled() )
								ObjectDisabled( derivedparent->AreChildsDisabled() );
							}
							else
								// do not detach here due to avoid changing the contents
								qWarning("YDerivedType<T>::SetParent() called with illegal arguments! That is real problem!");
						}
	//! The object will be detached from its parent. If it is currently not valid it will be set to the current value
	//! of its parent and a signal will be emitted. If the content is valid it will not change.
	//! If the object is disabled it will be enabled but all childs will be disabled.
	//! If the parent has disabled the childs the current object will take over this feature
	//! so alle kinds will be disabled further
	void				DetachParent()
						{
							if ( derivedparent )
							{
								T		old_content=GetContent();
								bool	old_disabled=disabled;
								bool	old_childsdisabled=childsdisabled;
								DisconnectDerivedParent(derivedparent);
								disabled=false;
								derivedparent=0;
								if ( IsValid() )
									SetContent(old_content);
								SetChildsDisabled( old_disabled || old_childsdisabled );
							}
						}

	//! returns true if this instance has a valid content
	virtual bool		IsValid() const						{	return valid;										}

	//! returns true if this instance is disabled. Disabled objects will return contents of the parent objects.
	bool				IsDisabled() const					{	return disabled;									}
	//! returns true if this instance has disabled its child objects
	bool				AreChildsDisabled() const			{	return disabled || childsdisabled;					}
	//! Sets this object and all child objects into disable state
	//! Disabled objects will return contents of the parent objects.
	void				SetChildsDisabled(bool b)			{	if (childsdisabled!=b)
																{
																	childsdisabled=b;
																	DisableChildObjects(b);
																}
															}

	//! This function will unset the current content. If parent is zero this function has no effect!
	void				Invalid()							{	if (derivedparent && valid)
																{
																	valid=false;
																	ObjectUpdated();
																}
															}

	//! Returns the first non disable and valid ptr to content.
	//! This function is parsing the
	//! instances until the first occurence of a valid content.
	const YDerivedType<T>*	GetValidContent() const			{	const YDerivedType<T>	*derivedtype=this;
																do
																{
																	if (! derivedtype->IsDisabled() )
																		break;
																	derivedtype=derivedtype->derivedparent;
																} while ( derivedtype && derivedtype->derivedparent );
																while ( derivedtype->derivedparent && !derivedtype->IsValid() )
																	derivedtype=derivedtype->derivedparent;
																return derivedtype;
															}

	const YDerivedType<T>*	GetDerivedParent() const		{	return derivedparent;								}
	//! returns the first valid content. This function is parsing the instances until
	//! the first occurence of a valid content. This content will be returned.
	//! Disabled objects will return contents of the parent objects.
	const T&			GetContent() const					{	return GetValidContent()->content;					}

	//! This function is used to set the content of the object. Attention: During process there will be
	//! called the virtual function CheckLimits() to ensure the limits of the objects. In this way
	//! it is possible that the arguments will be modified!
	const T&			SetContent(const T &c)				{	T newcontent;
																if ( CheckLimits(newcontent,c) )
																{
																	if ( (content!=newcontent) || (!valid) )
																	{
																		valid=true;
																		content=newcontent;
																		ObjectUpdated();
																	}
																}
																//else
																//	Invalid();		// check limits returns false->content invalid ?
																return content;
															}

	//! This operator copies the content of the current instance to the destination. Destination
	//! is a type of <T> instead YDerivedType. There is another global operator defined if you used
	//! the macro TYPEDEF_DERIVE(type, name) to define the YSettins type
	//\example QString name<<ysettingname;
	const T&			operator >> (T& dest) const			{	dest=GetContent();
																return dest;										}
	//! This operator sets the content of the instance
	const T&			operator = (const T &c)				{	return SetContent(c);								}

	//! This operator will set the content of the current object to "c". After calling this
	//! this function the current object is valid and set!
	const YDerivedType<T>& operator = (const YDerivedType<T> &c)
															{	operator =(c.GetContent() );
																return *this;										}
	//! This operator sets the content of the instance. It is similar to "="
	const T&			operator <<(const T &c)				{	return operator =(c);								}

	//! This operator compares the object with the <c> and returns true if it is not equal
	bool				operator !=(const YDerivedType<T> &c) const
															{	return GetContent() != c.GetContent();				}
	//! This operator compares the object with the c and returns true if it is not equal
	bool				operator !=(const T &c) const		{	return GetContent() != c;							}
	//! This operator compares the object with the <c> and returns true if it is equal
	bool				operator ==(const YDerivedType<T> &c) const
															{	return GetContent() == c.GetContent();				}
	//! This operator compares the object with the <c> and returns true if it is equal
	bool				operator ==(const T &c) const		{	return GetContent() == c;							}
	//! This conversion operator checks, if the content is valid
						operator bool()						{	return IsValid();									}

	/** Checks the limits of <source> and copies a final version to dest which will be used during SetContent()
	//\return true=content can be set
	//\return false=content can not be set
	**/
	virtual bool		CheckLimits(T& dest, const T &source) const
															{	dest=source;
																return true;										}
	//! This is a slot in the base class (tricky right?) to ensure that derived objects
	//! can receive signals from their parent objects.
	//! This functions is called every time an parent object is changed. This function will refer the
	//! signal to its children. Due to possible changes in "object limits" we call SetContent() in
	//! this function.
	virtual void		DerivedParentContentChanged()		{	/*	qWarning("Name:%s", GetName());	*/
																// if there is a valid content it must be set again
																// to check the content limits
																if (derivedparent)
																{
																	if ( IsValid() )
																	{
																		//Invalid();			// force UpdateObject for child objects
																		valid=false;		//! prevent UpdateObject twice
																		SetContent( content );	// check current content again, maybe limits have changed
																	}
																	else
																		ObjectUpdated();
																}
															}
	//! If the parent object has called the function SetChildsDisabled() all child objects
	//! will receive the signal DisableObjects() in this slot. The content of all child
	//! objects will be disabled.
	virtual void		ObjectDisabled(bool b)				{
																if ( disabled!=b)
																{
																	disabled=b;
																	DerivedParentContentChanged();
																	DisableChildObjects(b);		// inform child objects
																}
															}
};



#define Typedef_YDerivedObject(type, name)	typedef YDerivedType<type> name; inline const type &	operator << (type &dest, const name &source) {	return source.operator >>(dest); }




//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif																	// ifndef YASAGUI_OBJECT_INCLUDE
