//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// YASA timing handler declarations
//
// Project: Yasa 2
// Author : Jan Blumenthal
// Start  : 2002/02/28
// $Header: /sources/yasa/yasagui/timinghandler.h,v 1.4 2003/01/24 15:47:43 bj Exp $
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program (See the included file COPYING);
// if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef YASAGUI_TIMING_HANDLER_INCLUDE
#define	YASAGUI_TIMING_HANDLER_INCLUDE


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <qscrollview.h>
#include <qsplitter.h>
#include "timingsettings.h"
#include "timingpainter.h"
#include "settings.h"
#include "resulthandler.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YTimings;
class YTimingHandler;


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class is used to store information about the timing diagrams in the scrollview class.
//! This class is sorted in a list within the YTimingHandler and will be deleted automatically if removed from list
class YTimingInfo
{	//! Stores absolute coordinates of this timing diagram
	QPoint				position;
	//! ptr to timing diagram class
	YTimings			*timings;
public:					YTimingInfo() : timings(0)
						{}
						~YTimingInfo();
	YTimings *			GetTimings()								{	return timings;				}
	const QPoint &		GetPosition() const							{	return position;			}
	void				SetTimings(YTimings *t)						{	timings=t;					}
	void				SetPosition(const QPoint &p)				{	position=p;					}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is the scrollview class is providing a scrollview showing timing diagrams
class YTimingScrollView : public QScrollView
{	Q_OBJECT
	YTimingHandler		*timinghandler;
protected:
	QSize				size;										// current max. size of widgets, depends on inserted widgets
public:					YTimingScrollView(QWidget *parent) : QScrollView(parent), timinghandler(0)
						{
							setResizePolicy(QScrollView::AutoOneFit);
						}
	void				SetTimingHandler(YTimingHandler *th)		{	timinghandler=th;				}
	YTimingHandler*		GetTimingHandler() const					{	return timinghandler;			}
	virtual void		Resize()									{									}
	const QSize &		GetSize() const								{	return size;					}
protected:
    virtual void		drawContents(QPainter* p, int cx, int cy, int cw, int ch)=0;
public slots:
	virtual void		NewTimingInserted(YTimingInfo *ti)=0;
	virtual void		ViewScrolled(int posx, int posy);
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is the scrollview class is providing a scrollview showing timing diagrams
class YTimingDiagramScrollView : public YTimingScrollView
{	Q_OBJECT
	QPoint				currentmousepos;
	//! leftbuttonpos valid ?
	bool				leftbuttonpressed;
	//! position of mouse (global) at the moment the left button was pressed
	QPoint				leftbuttonpos;
public:					YTimingDiagramScrollView(QWidget *parent) : YTimingScrollView(parent),
						leftbuttonpressed(false)
						{
							QCursor		*cursor;
							setHScrollBarMode( QScrollView::AlwaysOn );
							if ( ! YApplicationFigures::CreateCursor(&cursor, YApplicationFigures::POINTER_HAND) && cursor )
							{
								setCursor( *cursor );
							}
							setFocusPolicy( QWidget::WheelFocus );
							setFocus();
//							setMouseTracking(true);
						}
	virtual void		Resize();
protected:
	virtual void		keyPressEvent(QKeyEvent *event );
    virtual void		drawContents(QPainter* p, int cx, int cy, int cw, int ch);
	virtual void		contentsMousePressEvent ( QMouseEvent *event )
						{
							QCursor		*cursor;
							setFocus();
							if ( event->button() == Qt::LeftButton )
							{
								leftbuttonpressed=true;
								leftbuttonpos=event->globalPos();
								if ( ! YApplicationFigures::CreateCursor(&cursor, YApplicationFigures::POINTER_HAND_PUSHED) && cursor )
								{
									setCursor( *cursor );
								}
							}
						}
	virtual void		contentsMouseReleaseEvent ( QMouseEvent *event )
						{
							if ( event->button() == Qt::LeftButton )
							{
								QCursor			*cursor;
								leftbuttonpressed=false;
								if ( ! YApplicationFigures::CreateCursor(&cursor, YApplicationFigures::POINTER_HAND) && cursor )
								{
									setCursor( *cursor );
								}
							}
						}
	virtual void		contentsMouseMoveEvent( QMouseEvent *event )
						{
						//	currentmousepos=event->pos();
							if ( leftbuttonpressed )
							{
								scrollBy( leftbuttonpos.x()-event->globalX(), leftbuttonpos.y()-event->globalY() );
								leftbuttonpos = event->globalPos();
							}

						}
public slots:
	virtual void		NewTimingInserted(YTimingInfo *ti);
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This is the scrollview class providing a scrollview showing names of timing diagrams
class YTimingNameScrollView : public YTimingScrollView
{	Q_OBJECT
public:					YTimingNameScrollView(QWidget *parent) : YTimingScrollView(parent)
						{
							setVScrollBarMode( QScrollView::AlwaysOff );
							setHScrollBarMode( QScrollView::AlwaysOn );
						}
protected:
    virtual void		drawContents(QPainter* p, int cx, int cy, int cw, int ch);
public slots:
	virtual void		NewTimingInserted(YTimingInfo *ti);
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! The YTimingHandler is handling the timing diagram instances in a list and provides information to the different
//! versions of the YTimingScrollViews
class YTimingHandler : public YResultHandler
{	Q_OBJECT
	QWidget				*parent;
	YPainter			ypainter;
	QList<YTimingInfo>	timinginfos;
	//! This variable counts the number of calls to function ViewScrolled to prevent recursions
	unsigned int		scrollcount;
	void				UpdateTimings();
	YASA_TIME			absmaxtime;							// abs max time of all diagrams
public:					YTimingHandler(QWidget *p) : parent(p), scrollcount(0), absmaxtime(0)
						{
							timinginfos.setAutoDelete(true);
						}
	//! Insert a timing diagram and append it to the end of the list. This is used during initialiazation of the scrollview.
	int					InsertTiming(YTimings *t);
	virtual void		ClearResults()									{	timinginfos.clear();	}
	//! This function must be called after inserting all timing diagrams
	virtual int			PostCreate();
	YPainter*			GetPainter()									{	return &ypainter;		}
	QList<YTimingInfo>*	GetTimingInfos()								{	return &timinginfos;	}
	void				ViewScrolled(YTimingScrollView *tsc, int posx, int posy);
	void				ZoomIn();
	void				ZoomOut();
	//! date has changed. Maybe new zoomfactor
	virtual void		UpdateViews() =0;
signals:
	void				NewTimingInserted(YTimingInfo *ti);
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! The YDiagramNameTimingHandler is showing the timing instances in two splitted scrollviews
//! (YTimingNameScrollView and YTimingDiagramScrollView)
class YDiagramNameTimingHandler : public YTimingHandler
{	Q_OBJECT
	YTimingNameScrollView	namescrollview;
	YTimingDiagramScrollView diagramscrollview;
public:					YDiagramNameTimingHandler(QSplitter *parent) : YTimingHandler(parent),
						namescrollview(parent),
						diagramscrollview(parent)
						{
							namescrollview.SetTimingHandler(this);
							diagramscrollview.SetTimingHandler(this);
							connect(this, SIGNAL( NewTimingInserted(YTimingInfo *) ), &namescrollview, SLOT( NewTimingInserted(YTimingInfo *) ) );
							connect(this, SIGNAL( NewTimingInserted(YTimingInfo *) ), &diagramscrollview, SLOT( NewTimingInserted(YTimingInfo *) ) );
							connect( &namescrollview, SIGNAL( contentsMoving(int, int) ), &diagramscrollview, SLOT( ViewScrolled(int, int ) ) );
							connect( &diagramscrollview, SIGNAL( contentsMoving(int, int) ), &namescrollview, SLOT( ViewScrolled(int, int ) ) );
							// correct size of splitters
							QValueList<int>	widthlist=parent->sizes();
							int				w=widthlist[0]+widthlist[1];
							widthlist[0]=w*1/5;
							widthlist[1]=w*4/5;
							parent->setSizes( widthlist );
						}
	//! date has changed. Maybe new zoomfactor
	virtual void		UpdateViews();
};



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif																			// ifndef YASAGUI_TIMING_HANDLER_INCLUDE
