//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// YASA timing diagram window declarations
//
// Project: Yasa 2
// Author : Jan Blumenthal
// Start  : 2002/02/28
// $Header: /sources/yasa/yasagui/timings.h,v 1.4 2003/01/24 15:47:43 bj Exp $
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program (See the included file COPYING);
// if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This file contains all information about timing diagram frames
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef YASAGUI_TIMINGS_INCLUDE
#define	YASAGUI_TIMINGS_INCLUDE


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <yasagui/yasagui.h>
#include "datatypes.h"
#include "timingsettings.h"
#include "timingparser.h"
#include "environment.h"
#include "project.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class YPainter;
class YLogAction;


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class is an abstract base class providing an interface for different timing diagrams
//! It acts like a widget. It must be inserted with InsertTiming()
class YTimings
{	int					width;								// width of timing diagram (only diagram)
	int					height;								// height -++-
	int					namewidth;							// width of name widget (left splitter window)
	int					nameheight;							// height -++-
	int					timingheight;						// height of one timing entry
	int					timingsheight;						// height of all timing entries ( timingheight*entries )
	int					entries;							// number of entries
	YASA_TIME			maxduration;						// max duration of this diagram
	QString				title;								// title of widget ("This is a timing diagram of project xx"
	QString				ordertitle;							// order of widget ("Task" or "Mutex" for left splitter window)
	QStringList			entrynames;							// list of entry names
	QFontMetrics		*fontmetrics;						// font info of used font
	YTimingParser		*timingparser;
	int					titlewidth;

	int					GetEntries() const													{	return entries;			}
	int					CalculateHeight();
	int					CalculateWidth(int zoomfactor);
	const QFontMetrics* GetFontMetrics() const												{	return fontmetrics;		}
	const QString &		GetTitle() const													{	return title;			}
	const QString &		GetOrderTitle() const												{	return ordertitle;		}
	const QStringList &	GetRowNames() const													{	return entrynames;		}
	int					GetTitleHeight() const
						{
							return YTIMING_SPACING_Y + fontmetrics->height() + YTIMING_SPACING_Y+2;
						}
	int					GetTimingsHeight() const											{	return timingsheight;	}
	int					GetTitleWidth() const												{	return titlewidth;		}
public:					YTimings() :
						width(0),
						height(0),
						namewidth(0),
						nameheight(0),
						timingheight(0),
						timingsheight(0),
						entries(0),
						maxduration(0),
						fontmetrics(0),
						timingparser(0),
						titlewidth(0)
						{ }
	virtual 			~YTimings()
						{
							if (fontmetrics)
								delete fontmetrics;
						}
	virtual int			InitInstance(const QFont &font, YPainter *painter)
						{	maxduration=CalculateMaxDuration();
							entries=GetMaxEntries();
							title=GetTimingTitle();
							ordertitle=GetOrderName();
							entrynames=CreateRowNames();
							fontmetrics=new QFontMetrics(font);
							timingparser=CreateTimingParser();
							if ( fontmetrics && timingparser )
							{
								timingparser->SetPainter(painter);
								timingparser->SetTimings(this);
								titlewidth=fontmetrics->width( GetTitle() );
								CalculateHeight();
								CalculateWidth(YTIMING_DEFAULT_ZOOMFACTOR);
								return 0;
							}
							else
								return YASAE_OUT_OF_MEMORY;
						}
	virtual void		UpdateInstance(int zoomfactorx)
						{	if ( fontmetrics->height() > YTIMING_TIMING_HEIGHT )
								timingheight=fontmetrics->height() + YTIMING_SPACING_Y;
							else
								timingheight=YTIMING_TIMING_HEIGHT + YTIMING_SPACING_Y;
							timingsheight=timingheight * GetEntries();
							CalculateHeight( );
							CalculateWidth( zoomfactorx );
						}
	int					GetTimingHeight() const										{	return timingheight;	}
	int					GetHeight() const											{	return height;		}
	int					GetWidth() const											{	return width;		}
	int					GetNameHeight() const										{	return nameheight;	}
	int					GetNameWidth() const										{	return namewidth;	}
    virtual void		DrawTimingContents(YPainter& p) const;
    virtual void		DrawNameContents(YPainter& p) const;
	YASA_TIME			GetMaxDuration() const										{	return maxduration;	}
	//! This function returns the number of the row of the given logging action. This
	//! function should be derived if needed.
	virtual int			GetLogActionRow(const YLogAction *) const					{	return 0;			}
	//! Filter this logging action or not
	virtual bool		FilterLoggingAction(const YLogAction *) const				{	return true;		}
protected:
	//! Returns the number of rows in the timing widget
	virtual unsigned int GetMaxEntries() const=0;
	//! Returns a list of names which should be drawn in each row
	virtual QStringList	CreateRowNames() const=0;
	//! Returns the order name "Task" or "Mutex" or "Environment" used to create the title string
	//! Example: "Timing diagram of task 1 of project - CPU order"
	virtual QString		GetOrderName() const=0;
	virtual QString		GetTimingTitle() const=0;
	virtual YASA_TIME	CalculateMaxDuration() const=0;
	virtual YTimingParser * CreateTimingParser() const=0;
	//! Converts the given time into coordinates within the YTimingRect
	int					ConvertTimeToPosition(YPainter &p, YASA_TIME time) const	{	return time / p.GetZoomFactor();}
	YTimingParser *		GetTimingParser() const										{	return timingparser;			}
	virtual void		EvaluateTimings() const	=0;
	//! Returns the string which should be printed in the drawn rectangle, e.g. "Task 1"
	virtual QString		GetDurationString(const YLogAction *) const					{	return QString::null;			}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class can handle all timing diagrams which are using one YEnvironment of the project
class YEnvironmentTimings : public YTimings
{	protected:
	YEnvironmentLogFileSet &environmentlogfile;
public:					YEnvironmentTimings(YEnvironmentLogFileSet &e) : environmentlogfile(e)			{ }
	virtual QString		GetTimingTitle() const
						{
							YEnvironment *environment=(YEnvironment*) environmentlogfile.GetParent();
							return QString( MSGP_TI_TITLE_ENVIRONMENT )
											.arg( MSGP_TI_SUBTITLE_PROJECT )
											.arg( environment->GetProject()->GetName() )
											.arg( environment->GetName() )
											.arg( GetOrderName() );
						}
	virtual void		EvaluateTimings() const				{	environmentlogfile.EvaluateLoggingActions( *GetTimingParser() );	}
	virtual YASA_TIME	CalculateMaxDuration() const		{	return environmentlogfile.GetMaxDuration();							}
	//! Returns the string which should be printed in the drawn rectangle
	//! Default return value: "Taskname" if more than task is defined
	virtual QString		GetDurationString(const YLogAction *action) const
	{
		if ( environmentlogfile.GetTaskSet().GetCount() )		// does this logfile have more than one task ?
		{
			YLogFileTask *thread=action->GetThread();
			if ( thread )
				return thread->GetName();
		}
		return QString::null;
	}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows the task timing results of the given environment
//!
//! Task 1 | XXXXX           XXXXXX
//! Task 2 |      XXX    XXXX      XXXX
//! Task 3 |         XXXX              XXXXX
//!       +--------------------------------------------------------
class YTaskEnvironmentTimings : public YEnvironmentTimings
{public:				YTaskEnvironmentTimings(YEnvironmentLogFileSet &e) : YEnvironmentTimings(e)		{ }
	virtual QStringList	CreateRowNames() const
						{	QStringList	list;
							for ( unsigned int i=0; i<GetMaxEntries() ; i++ )
								list.append( environmentlogfile.GetTaskSet().At(i)->GetName() );
							return list;
						}
	virtual unsigned int GetMaxEntries() const				{	return environmentlogfile.GetTaskSet().GetCount();		}
	virtual QString		GetOrderName() const				{	return MSGP_TI_SUBTITLE_TASK;							}
	virtual YTimingParser * CreateTimingParser() const		{	return new YTaskEnvironmentTimingParser();							}
	//! Return number of cpu or QString::null
	virtual QString		GetDurationString(const YLogAction *) const
	{
		if ( environmentlogfile.GetCount() )			// does this logfile have more than one cpu ?
			return QString::number( GetTimingParser()->GetCPULogFile()->GetCPU() );
		return QString::null;
	}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows the usage of mutexes: The printed numbers are number of the using task
//!
//! Mutex 1 | XX1XX  XX2XX XXX3XXXX XXXX3XXXX
//! Mutex 2 |    XX1XX   XX1XX      XX4XX
//! Mutex 3 |        XX3XX   XXX2XXXX
//!       +--------------------------------------------------------
class YMutexEnvironmentTimings : public YEnvironmentTimings
{public:				YMutexEnvironmentTimings(YEnvironmentLogFileSet &e) : YEnvironmentTimings(e)				{ }
	virtual QStringList	CreateRowNames() const
						{	QStringList	list;
							for ( unsigned int i=0; i<GetMaxEntries(); i++ )
								list.append( environmentlogfile.GetMutexSet().At(i)->GetName() );
							return list;
						}
	virtual unsigned int GetMaxEntries() const						{	return environmentlogfile.GetMutexSet().GetCount();		}
	virtual QString		GetOrderName() const						{	return MSGP_TI_SUBTITLE_MUTEX;				}
	virtual YTimingParser * CreateTimingParser() const				{	return new YMutexEnvironmentTimingParser();		}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows the CPU timing results of the given environment
//!
//! CPU 1 | (--1--)(--2--)  (-----3-----)
//! CPU 2 |     (----4----)(------5------)(6)
//! CPU 3 | (7)         (--8--)
//!       +--------------------------------------------------------
class YCPUEnvironmentTimings : public YEnvironmentTimings
{public:				YCPUEnvironmentTimings(YEnvironmentLogFileSet &e) : YEnvironmentTimings(e)					{ }
	virtual QStringList	CreateRowNames() const
						{	QStringList	list;
							for ( unsigned int i=0; i<GetMaxEntries(); i++ )
								list.append( environmentlogfile.At(i)->GetName() );
							return list;
						}
	virtual unsigned int GetMaxEntries() const							{	return environmentlogfile.GetCount();	}
	virtual QString		GetOrderName() const							{	return MSGP_TI_SUBTITLE_CPU;			}
	virtual YTimingParser * CreateTimingParser() const					{	return new YCPUEnvironmentTimingParser();	}
};



















//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This class can handle all environments of the whole projects
class YProjectTimings : public YTimings
{	protected:
	YProjectLogFile&	projectlogfile;
public:					YProjectTimings(YProjectLogFile &lfs) : projectlogfile(lfs)		{ }
	virtual void		EvaluateTimings() const						{	projectlogfile.EvaluateLoggingActions( *GetTimingParser() );}
	virtual QString		GetTimingTypeName() const=0;
	virtual YASA_TIME	CalculateMaxDuration() const				{	return projectlogfile.GetMaxDuration();		}
	virtual QStringList	CreateRowNames() const						{	return projectlogfile.GetEnvironmentNames();}
	virtual QString		GetOrderName() const						{	return MSGP_TI_SUBTITLE_ENVIRONMENT;		}
	virtual unsigned int GetMaxEntries() const						{	return projectlogfile.GetMaxEntries();		}
	virtual int			GetLogActionRow(const YLogAction *) const	{	return GetTimingParser()->GetEnvironmentLogFileNumber();		}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows the task timing results of in all environments
//!
//! Task "Swapping" of project "This is a demo project" - environment order
//! Simulator (LLF)  | XX XX XX XX XX XX XX XX XX XX XX
//! Simulator (ELLF) | XXXXXXXXXX            XXXXXXXXXX
//! RT-Linux (ELLF)  | XXXX                  XXXXXX
//!                  +--------------------------------------------------------
class YTaskProjectTimings : public YProjectTimings
{	int					taskid;
	QString				taskname;
public:					YTaskProjectTimings(const QString &tname, int id, YProjectLogFile &p) :
						YProjectTimings(p), taskid(id), taskname(tname)												{ }
	virtual YTimingParser * CreateTimingParser() const				{	return new YTaskProjectTimingParser();		}
	virtual QString		GetTimingTypeName() const					{	return MSGP_TI_SUBTITLE_TASK;				}
	int					GetTaskID() const							{	return taskid;								}
	virtual bool		FilterLoggingAction(const YLogAction *action) const
																	{	return taskid == action->GetThread()->GetId();	}
	virtual QString		GetTimingTitle() const
						{	return QString(MSGP_TI_TITLE_PROJECT)
											.arg( GetTimingTypeName() )
											.arg( taskname )
											.arg( GetOrderName() );
						}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows the mutex timing results in all environments
//!
//! Mutex "IO-Address" of project "This is a demo project" - environment order
//! Simulator (LLF)  |          XXXXXXX
//! Simulator (ELLF) |            XXX
//! RT-Linux (ELLF)  |             X
//!                  +--------------------------------------------------------
class YMutexProjectTimings : public YProjectTimings
{	int					mutexid;
	QString				mutexname;
public:					YMutexProjectTimings(const QString &mname, int id, YProjectLogFile &p) : YProjectTimings(p),
						mutexid(id), mutexname(mname)															{ }
	virtual YTimingParser * CreateTimingParser() const				{	return new YMutexProjectTimingParser();	}
	virtual QString		GetTimingTypeName() const					{	return MSGP_TI_SUBTITLE_MUTEX;			}
	int					GetMutexID() const							{	return mutexid;							}
	virtual bool		FilterLoggingAction(const YLogAction *action) const
																	{	return action->GetMutex() && action->GetMutex()->GetId()==mutexid;	}
	virtual QString		GetTimingTitle() const
						{	return QString(MSGP_TI_TITLE_PROJECT)
											.arg( GetTimingTypeName() )
											.arg( mutexname )
											.arg( GetOrderName() );
						}
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//! This frame shows one CPU timing results of all environment
//!
//! Timing diagram of CPU 0 - environment order
//! Simulator (LLF)  |  XXXXXXXXXXXXXXXXXXXXX
//! Simulator (ELLF) |  XXXXXXXXX             XXXXXXXXX
//! RT-Linux (ELLF)  |  XXX                   XXXXX
//!                  +--------------------------------------------------------
class YCPUProjectTimings : public YProjectTimings
{	unsigned int		cpu;
public:					YCPUProjectTimings(unsigned int c, YProjectLogFile &p) : YProjectTimings(p), cpu (c)				{ }
	virtual YTimingParser * CreateTimingParser() const				{	return new YCPUProjectTimingParser();			}
	virtual QString		GetTimingTypeName() const					{	return MSGP_TI_SUBTITLE_CPU;			}
	unsigned int		GetCPU() const								{	return cpu;								}
	//virtual bool		FilterLoggingAction(const YLogAction *action) const
	//																{	return action->GetMutex() && action->GetMutex()->GetId()==mutexid;	}
	virtual QString		GetTimingTitle() const
						{	return QString(MSGP_TI_TITLE_PROJECT)
											.arg( GetTimingTypeName() )
											.arg( cpu )
											.arg( GetOrderName() );
						}
};
















//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif																			// ifndef YASAGUI_TIMINGS_INCLUDE
